# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
A pattern which models the occurrence of a phenomenon and facilitates the
generating of a complex event.
"""

from typing import Tuple, List

from bobocep import BoboError
from bobocep.cep.phenom.pattern.predicate import BoboPredicate

_EXC_PREDICATES_LEN = "predicates must have a length greater than 0"
_EXC_STRICT_OPT_TRUE = "strict and optional must not both be True"
_EXC_NEG_OR_OPT_LOOP_TRUE = "negated and optional must " \
                            "both be False if loop is True"
_EXC_NEG_AND_OPT_LOOP_FALSE = "negated and optional must not " \
                              "both be True if loop is False"

_EXC_NAME_LEN = "name must have a length greater than 0"
_EXC_BLOCKS_LEN = "blocks must have a length greater than 0"
_EXC_BLOCK_FIRST_NOT_NEG = "first block cannot be negated"
_EXC_BLOCK_FIRST_NOT_OPT = "first block cannot be optional"
_EXC_BLOCK_FIRST_NOT_LOOP = "first block cannot loop"
_EXC_BLOCK_LAST_NOT_NEG = "last block cannot be negated"
_EXC_BLOCK_LAST_NOT_OPT = "last block cannot be optional"
_EXC_BLOCK_LAST_NOT_LOOP = "last block cannot loop"


class BoboPatternError(BoboError):
    """
    A pattern error.
    """


class BoboPatternBlockError(BoboPatternError):
    """
    A pattern block error.
    """


class BoboPatternBlock:
    """
    A pattern block.
    """

    def __init__(self,
                 predicates: List[BoboPredicate],
                 group: str,
                 strict: bool,
                 loop: bool,
                 negated: bool,
                 optional: bool):
        """
        :param predicates: Block predicates.
        :param group: The group with which the block is associated.
            Can be an empty string.
        :param strict: `True` if the block has strict contiguity;
            `False` otherwise.
        :param loop: `True` if the block loops back to itself;
            `False` otherwise.
        :param negated: `True` if the block is negated;
            `False` otherwise.
        :param optional: `True` if the block is optional;
            `False` otherwise.
        """
        super().__init__()

        if len(predicates) == 0:
            raise BoboPatternBlockError(_EXC_PREDICATES_LEN)

        if strict and optional:
            raise BoboPatternBlockError(_EXC_STRICT_OPT_TRUE)

        if loop and (negated or optional):
            raise BoboPatternBlockError(_EXC_NEG_OR_OPT_LOOP_TRUE)

        if (not loop) and (negated and optional):
            raise BoboPatternBlockError(_EXC_NEG_AND_OPT_LOOP_FALSE)

        self._predicates: Tuple[BoboPredicate, ...] = tuple(predicates)
        self._group: str = group
        self._strict: bool = strict
        self._loop: bool = loop
        self._negated: bool = negated
        self._optional: bool = optional

    @property
    def predicates(self) -> Tuple[BoboPredicate, ...]:
        """
        :return: Block predicates.
        """
        return self._predicates

    @property
    def group(self) -> str:
        """
        :return: Block group.
        """
        return self._group

    @property
    def strict(self) -> bool:
        """
        :return: `True` if pattern block has strict contiguity;
            `False` otherwise.
        """
        return self._strict

    @property
    def loop(self) -> bool:
        """
        :return: `True` if pattern block loops;
            `False` otherwise.
        """
        return self._loop

    @property
    def negated(self) -> bool:
        """
        :return: `True` if pattern block is negated;
            `False` otherwise.
        """
        return self._negated

    @property
    def optional(self) -> bool:
        """
        :return: `True` if pattern block is optional;
            `False` otherwise.
        """
        return self._optional


class BoboPattern:
    """
    A pattern that represents a means by which to detect the occurrence of
    some phenomenon.
    """

    def __init__(self,
                 name: str,
                 blocks: List[BoboPatternBlock],
                 preconditions: List[BoboPredicate],
                 haltconditions: List[BoboPredicate]):
        """
        :param name: The pattern name.
        :param blocks: The pattern blocks.
        :param preconditions: The pattern preconditions.
        :param haltconditions: The pattern haltconditions.
        """
        super().__init__()

        if len(name) == 0:
            raise BoboPatternError(_EXC_NAME_LEN)

        if len(blocks) == 0:
            raise BoboPatternError(_EXC_BLOCKS_LEN)

        if blocks[0].negated:
            raise BoboPatternError(_EXC_BLOCK_FIRST_NOT_NEG)

        if blocks[0].optional:
            raise BoboPatternError(_EXC_BLOCK_FIRST_NOT_OPT)

        if blocks[0].loop:
            raise BoboPatternError(_EXC_BLOCK_FIRST_NOT_LOOP)

        if blocks[-1].negated:
            raise BoboPatternError(_EXC_BLOCK_LAST_NOT_NEG)

        if blocks[-1].optional:
            raise BoboPatternError(_EXC_BLOCK_LAST_NOT_OPT)

        if blocks[-1].loop:
            raise BoboPatternError(_EXC_BLOCK_LAST_NOT_LOOP)

        self._name: str = name
        self._blocks: Tuple[BoboPatternBlock, ...] = tuple(blocks)
        self._preconditions: Tuple[BoboPredicate, ...] = tuple(preconditions)
        self._haltconditions: Tuple[BoboPredicate, ...] = tuple(haltconditions)

    @property
    def name(self) -> str:
        """
        :return: Pattern name.
        """
        return self._name

    @property
    def blocks(self) -> Tuple[BoboPatternBlock, ...]:
        """
        :return: Pattern blocks.
        """
        return self._blocks

    @property
    def preconditions(self) -> Tuple[BoboPredicate, ...]:
        """
        :return: Pattern preconditions.
        """
        return self._preconditions

    @property
    def haltconditions(self) -> Tuple[BoboPredicate, ...]:
        """
        :return: Pattern haltconditions.
        """
        return self._haltconditions



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
CEP engine tasks.
"""

from abc import ABC, abstractmethod

from bobocep import BoboError


class BoboEngineTaskError(BoboError):
    """
    An engine task error.
    """


class BoboEngineTask(ABC):
    """
    An engine task.
    """

    @abstractmethod
    def update(self) -> bool:
        """
        :return: `True` if an update occurred in the task; `False` otherwise.
        """

    @abstractmethod
    def size(self) -> int:
        """
        :return: The number of events that the task is currently handling.
        """

    @abstractmethod
    def close(self) -> None:
        """
        Closes the engine task.
        """

    @abstractmethod
    def is_closed(self) -> bool:
        """
        :return: `True` if task is set to close; `False` otherwise.
        """



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Event history.
"""

from json import dumps, loads
from typing import Dict, List, Optional, Tuple

from bobocep.bobocep import BoboJSONable
from bobocep.cep.event.event import BoboEvent


class BoboHistory(BoboJSONable):
    """
    An event history.
    """

    def __init__(self, events: Dict[str, List[BoboEvent]]):
        """
        :param events: The history of events.
            Keys are group names.
            Values are lists of BoboEvent instances associated with a group.
        """
        super().__init__()

        self._events: Dict[str, List[BoboEvent]] = {}
        self._first: Optional[BoboEvent] = None
        self._last: Optional[BoboEvent] = None

        if events is not None:
            for name, event_list in events.items():
                for event in event_list:
                    if name not in self._events:
                        self._events[name] = []

                    self._events[name].append(event)

                    if self._first is None or \
                            event.timestamp < self._first.timestamp:
                        self._first = event

                    if self._last is None or \
                            event.timestamp > self._last.timestamp:
                        self._last = event

    @property
    def events(self) -> Dict[str, List[BoboEvent]]:
        """
        :return: All history events, indexed by group.
        """
        eventscopy: Dict[str, List[BoboEvent]] = {}

        for grp in self._events.keys():
            if grp not in eventscopy:
                eventscopy[grp] = []

            eventscopy[grp] = [eve for eve in self._events[grp]]

        return eventscopy

    def size(self) -> int:
        """
        :return: The total number of history events across all groups.
        """
        count = 0

        for grp in self._events.keys():
            count += len(self._events[grp])

        return count

    def all_groups(self) -> Tuple[str, ...]:
        """
        :return: All history groups in a tuple.
        """
        return tuple(self._events.keys())

    def all_events(self) -> Tuple[BoboEvent, ...]:
        """
        :return: All history events in a tuple.
        """
        all_events = []
        for grp in self._events.keys():
            all_events += self._events[grp]
        return tuple(all_events)

    def group(self, group: str) -> Tuple[BoboEvent, ...]:
        """
        :param group: A group name.
        :return: The BoboEvent instances associated with `group`.
        """
        if group in self._events:
            return tuple(self._events[group])
        else:
            return tuple()

    def first(self) -> Optional[BoboEvent]:
        """
        :return: The BoboEvent with the oldest timestamp,
            if there is at least one BoboEvent in the history.
        """
        return self._first

    def last(self) -> Optional[BoboEvent]:
        """
        :return: The BoboEvent with the most recent timestamp,
            if there is at least one BoboEvent in the history.
        """
        return self._last

    def to_json_dict(self) -> dict:
        """
        :return: A JSON `dict` representation of the history.
        """
        d: Dict[str, List[BoboEvent]] = {}

        for key in self._events:
            d[key] = [e for e in self._events[key]]

        return d

    def to_json_str(self) -> str:
        """
        :return: A JSON `str` representation of the history.
        """
        return dumps(self.to_json_dict(), default=lambda o: o.to_json_str())

    @staticmethod
    def from_json_str(j: str) -> 'BoboHistory':
        """
        :param j: A JSON `str` representation of the history.
        :return: A new instance of the history.
        """
        return BoboHistory.from_json_dict(loads(j))

    @staticmethod
    def from_json_dict(d: dict) -> 'BoboHistory':
        """
        :param d: A JSON `dict` representation of the history.
        :return: A new instance of the history.
        """
        from bobocep.cep.event.factory import BoboEventFactory

        events: Dict[str, List[BoboEvent]] = {}

        for key in d:
            events[key] = [BoboEventFactory.from_json_str(e) for e in d[key]]

        return BoboHistory(events=events)

    def __str__(self) -> str:
        """
        :return: A JSON `str` representation of the history.
        """
        return self.to_json_str()



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Engine task that forwards actions to external sources and generates
action events.
"""

from queue import Queue
from threading import RLock
from typing import Dict, List, Optional

from bobocep.cep.action.handler import BoboActionHandler, BoboHandlerResponse
from bobocep.cep.engine.forwarder.pubsub import BoboForwarderPublisher, \
    BoboForwarderSubscriber
from bobocep.cep.engine.producer.pubsub import BoboProducerSubscriber
from bobocep.cep.engine.task import BoboEngineTaskError, BoboEngineTask
from bobocep.cep.event import BoboEventAction, BoboEventComplex
from bobocep.cep.gen import BoboGenTimestamp
from bobocep.cep.gen.event_id import BoboGenEventID
from bobocep.cep.phenom.phenom import BoboPhenomenon

_EXC_PHENOM_NAME_DUP = "duplicate name in phenomena: {}"
_EXC_QUEUE_FULL = "queue is full (max size: {})"


class BoboForwarderError(BoboEngineTaskError):
    """
    A forwarder task error.
    """


class BoboForwarder(BoboEngineTask,
                    BoboForwarderPublisher,
                    BoboProducerSubscriber):
    """
    A forwarder task.
    """

    def __init__(self,
                 phenomena: List[BoboPhenomenon],
                 handler: BoboActionHandler,
                 gen_event_id: BoboGenEventID,
                 gen_timestamp: BoboGenTimestamp,
                 local_only: bool = True,
                 max_size: int = 0):
        """
        :param phenomena: List of phenomena.
        :param handler: Action handler.
        :param gen_event_id: Event ID generator.
        :param gen_timestamp: Timestamp generator.
        :param local_only: If `True`, forwarder only executes actions for
            locally-completed complex events i.e. not complex events that
            were completed on a distributed instance. **Note**: if `False`,
            the action may be executed more than once: once by the instance
            that generated the remote event, and another by this instance.
        :param max_size: Maximum queue size.
            Default: 0 (unbounded).
        """
        super().__init__()

        self._lock: RLock = RLock()
        self._closed: bool = False
        self._subscribers: List[BoboForwarderSubscriber] = []

        self._phenomena: Dict[str, BoboPhenomenon] = {}

        for phenom in phenomena:
            if phenom.name not in self._phenomena:
                self._phenomena[phenom.name] = phenom
            else:
                raise BoboForwarderError(
                    _EXC_PHENOM_NAME_DUP.format(phenom.name))

        self._handler: BoboActionHandler = handler
        self._gen_event_id: BoboGenEventID = gen_event_id
        self._gen_timestamp: BoboGenTimestamp = gen_timestamp
        self._max_size: int = max(0, max_size)
        self._local_only: bool = local_only
        self._queue: Queue[BoboEventComplex] = Queue(self._max_size)

    def subscribe(self, subscriber: BoboForwarderSubscriber):
        """
        :param subscriber: Subscriber to Forwarder data.
        """
        with self._lock:
            if subscriber not in self._subscribers:
                self._subscribers.append(subscriber)

    def update(self) -> bool:
        """
        :return: `True` if an internal update occurred; `False` otherwise.
        """
        with self._lock:
            if self._closed:
                return False

            handle = self._update_handler()
            response = self._update_responses()
            return handle or response

    def close(self) -> None:
        """
        Closes the Forwarder.
        """
        with self._lock:
            self._closed = True

    def is_closed(self) -> bool:
        """
        :return: `True` if Forwarder is closed; `False` otherwise.
        """
        with self._lock:
            return self._closed

    def _update_handler(self) -> bool:
        """
        :return: `True` if update occurred; `False` otherwise.
        """
        if not self._queue.empty():
            event: BoboEventComplex = self._queue.get_nowait()

            if event.phenomenon_name in self._phenomena:
                phenom: BoboPhenomenon = \
                    self._phenomena[event.phenomenon_name]

                if phenom.action is not None:
                    self._handler.handle(
                        action=phenom.action,
                        event=event)
            return True
        return False

    def _update_responses(self) -> bool:
        """
        :return: `True` if subscribers were notified of an action event
            from the action handler; `False` otherwise.
        """
        hres: Optional[BoboHandlerResponse] = \
            self._handler.get_handler_response()

        if hres is not None:
            # Generate action event from handler response
            event = BoboEventAction(
                event_id=self._gen_event_id.generate(),
                timestamp=self._gen_timestamp.generate(),
                data=hres.data,
                phenomenon_name=hres.complex_event.phenomenon_name,
                pattern_name=hres.complex_event.pattern_name,
                action_name=hres.action_name,
                success=hres.success)

            for subscriber in self._subscribers:
                subscriber.on_forwarder_update(event)

            return True

        return False

    def on_producer_update(
            self,
            event: BoboEventComplex,
            local: bool
    ) -> None:
        """
        :param event: Complex event generated by Producer.
        :param local: `True` if the complex event was generated using
            a locally-completed run; `False` otherwise.
        """
        with self._lock:
            if self._closed:
                return

            # Forwarder set to only execute actions for locally-generated
            # complex events.
            if (not local) and self._local_only:
                return

            if not self._queue.full():
                self._queue.put(event)
            else:
                raise BoboForwarderError(
                    _EXC_QUEUE_FULL.format(self._max_size))

    def size(self) -> int:
        """
        :return: Queue size.
        """
        with self._lock:
            return self._queue.qsize()



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Tools to help with `BoboCEP` setup.
"""

from abc import ABC, abstractmethod
from typing import Any

from bobocep import BoboError


class BoboSetupError(BoboError):
    """
    A setup error.
    """


class BoboSetup(ABC):
    """
    A setup.
    """

    @abstractmethod
    def generate(self) -> Any:
        """
        :return: Relevant setup data.
        """



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Distributed `BoboCEP` via TCP.
"""

import json
import logging
import socket
import time
from queue import Queue
from threading import Thread, RLock
from typing import Dict, Tuple, Optional, List

from bobocep.bobocep import BoboJSONableError, BoboJSONable
from bobocep.cep.engine.decider.pubsub import BoboDeciderSubscriber, \
    BoboDeciderPublisher
from bobocep.cep.engine.decider.runserial import BoboRunSerial
from bobocep.dist.crypto.crypto import BoboDistributedCrypto
from bobocep.dist.device import BoboDevice
from bobocep.dist.devman import BoboDeviceManager
from bobocep.dist.dist import BoboDistributedJSONDecodeError, \
    BoboDistributed, BoboDistributedError, BoboDistributedSystemError, \
    BoboDistributedTimeoutError
from bobocep.dist.pubsub import BoboDistributedSubscriber, \
    BoboDistributedPublisher

_KEY_COMPLETED = "completed"
_KEY_HALTED = "halted"
_KEY_UPDATED = "updated"

_TYPE_SYNC = 0
_TYPE_PING = 1
_TYPE_RESYNC = 2

_FLAG_RESET = 1

_EXC_CLOSED = "distributed is closed"
_EXC_RUNNING = "distributed is already running"
_EXC_NOT_RUNNING = "distributed is not running"


class _OutgoingJSONEncoder(json.JSONEncoder):
    """
    JSON encoder for outgoing messages.
    """

    def __init__(self, *args, **kwargs):
        """
        Constructor for encoding outgoing data into JSON.

        :param args: Arguments.
        :param kwargs: Keyword arguments.
        """
        json.JSONEncoder.__init__(self, *args, **kwargs)

    def default(self, obj: BoboJSONable):
        """
        :param obj: A JSONable object.
        :return: A JSON string.
        """
        try:
            return obj.to_json_str()

        except (KeyError, RecursionError, TypeError, ValueError) as e:
            raise BoboDistributedJSONDecodeError(
                "Failed to encode outgoing data. "
                "Ensure all types are JSONable. "
                "Error: {}".format(e))


class _IncomingJSONDecoder(json.JSONDecoder):
    """
    A JSON decoder for incoming messages.
    """

    def __init__(self):
        """
        Constructor for decoding incoming JSON data.
        """

        json.JSONDecoder.__init__(self, object_hook=self.object_hook)

    def object_hook(self, d: dict) -> Dict[str, List[BoboRunSerial]]:
        """
        :param d: An object dict.
        :return: Incoming Decider update information.
        """

        try:
            if _KEY_COMPLETED in d:
                d[_KEY_COMPLETED] = [
                    BoboRunSerial.from_json_str(rt)
                    for rt in d[_KEY_COMPLETED]]

            if _KEY_HALTED in d:
                d[_KEY_HALTED] = [
                    BoboRunSerial.from_json_str(rt)
                    for rt in d[_KEY_HALTED]]

            if _KEY_UPDATED in d:
                d[_KEY_UPDATED] = [
                    BoboRunSerial.from_json_str(rt)
                    for rt in d[_KEY_UPDATED]]

            return d

        except (KeyError, RecursionError, TypeError, ValueError) as e:
            # json.JSONDecodeError is a subclass of ValueError
            raise BoboDistributedJSONDecodeError(
                "Failed to decode incoming data. "
                "Ensure all types are JSONable."
                "Error: {}".format(e))


class BoboDistributedTCP(BoboDistributed,
                         BoboDistributedPublisher,
                         BoboDeciderSubscriber):
    """
    An implementation of distributed BoboCEP that uses TCP for data
    transmission across the network.
    """

    def __init__(self,
                 urn: str,
                 decider: BoboDeciderPublisher,
                 devices: List[BoboDevice],
                 crypto: BoboDistributedCrypto,
                 max_size_incoming: int = 0,
                 max_size_outgoing: int = 0,
                 period_ping: int = 30,
                 period_resync: int = 60,
                 attempt_stash: int = 5,
                 attempt_ping: int = 5,
                 attempt_resync: int = 10,
                 max_listen: int = 3,
                 timeout_accept: int = 3,
                 timeout_connect: int = 3,
                 timeout_send: int = 3,
                 timeout_receive: int = 3,
                 recv_bytes: int = 2048,
                 flag_reset: bool = True):
        """
        :param urn: A URN that is unique across devices in the network.
        :param decider: The Decider used in the local engine.
        :param devices: Devices in the network (including this device).
        :param crypto: Encryption to use for message exchange.
        :param max_size_incoming: Max queue size for incoming data.
            Default: 0 (unbounded).
        :param max_size_outgoing: Max queue size for outgoing data.
            Default: 0 (unbounded).
        :param period_ping: Period of inactivity from another device
            to warrant pinging the device, in seconds.
            Default: 30.
        :param period_resync: Period of inactivity from another device
            to warrant resyncing with the device, in seconds.
            Default: 60.
        :param attempt_stash: How frequently to attempt to send the sync stash
            if the stash is not empty.
        :param attempt_ping: How frequently to ping another device
            if it is within the ping period, in seconds.
            Default: 10.
        :param attempt_resync: How frequently to resync with another device
            if it is within the resync period, in seconds.
            Default: 10.
        :param max_listen: Max number of incoming connections to listen for
            at a given time, in seconds.
            Default: 3.
        :param timeout_accept: Timeout for accepting a new incoming
            connection, in seconds.
            Default: 3.
        :param timeout_connect: Timeout for connecting to a client
            when sending data, in seconds.
            Default: 3.
        :param timeout_send: Timeout for sending data, in seconds.
            Default: 3.
        :param timeout_receive: Timeout for receiving data, in seconds.
            Default: 3.
        :param recv_bytes: Number of bytes to receive at a time when receiving
            data.
            Default: 2048.
        :param flag_reset: If `True`, the RESET flag is set to indicate to
            external devices that it should reset its data on this device,
            which will trigger a resync.
        """
        super().__init__()

        # Lock used for local updates to and from the decider
        self._lock_local: RLock = RLock()
        # Lock used when sending and receiving remote messages
        self._lock_in_out: RLock = RLock()

        self._closed: bool = False
        self._running: bool = False
        self._thread_closed: bool = False
        self._subscribers: List[BoboDistributedSubscriber] = []

        self._urn: str = urn
        self._decider: BoboDeciderPublisher = decider
        self._devices: Dict[str, BoboDeviceManager] = {}
        self._crypto: BoboDistributedCrypto = crypto

        self._period_ping: int = period_ping
        self._period_resync: int = period_resync
        self._attempt_stash: int = attempt_stash
        self._attempt_ping: int = attempt_ping
        self._attempt_resync: int = attempt_resync

        # Check devices for duplicate URNs
        for d in devices:
            if d.urn in self._devices:
                raise BoboDistributedError(
                    "Duplicate device URN: {}".format(d.urn))

            self._devices[d.urn] = BoboDeviceManager(
                device=d,
                flag_reset=flag_reset)

        # Ensure at least two devices are listed, one of which is this device
        if (len(self._devices.keys()) < 2) or (urn not in self._devices):
            raise BoboDistributedError(
                "Devices list must contain at least 2 devices, one of which"
                " is this device ({}), found: {}"
                .format(urn, list(self._devices.keys())))

        self._max_listen: int = max_listen
        self._timeout_accept: int = timeout_accept
        self._timeout_connect: int = timeout_connect
        self._timeout_send: int = timeout_send
        self._timeout_receive: int = timeout_receive
        self._recv_bytes: int = recv_bytes

        self._thread_incoming: Thread = Thread(target=self._tcp_incoming)
        self._thread_outgoing: Thread = Thread(target=self._tcp_outgoing)

        self._queue_incoming: Queue[Dict[str, List[BoboRunSerial]]] = \
            Queue(maxsize=max_size_incoming)
        self._queue_outgoing: Queue[Dict[str, List[BoboRunSerial]]] = \
            Queue(maxsize=max_size_outgoing)

    def run(self) -> None:
        """
        Runs the distributed instance.
        """
        with self._lock_local:
            if self._closed:
                raise BoboDistributedError(_EXC_CLOSED)

            if self._running:
                raise BoboDistributedError(_EXC_RUNNING)

            self._thread_incoming.start()
            self._thread_outgoing.start()

            self._running = True

        while True:
            with self._lock_local:
                if self._closed:
                    self._running = False
                    break
                self._update()

    def subscribe(self, subscriber: BoboDistributedSubscriber) -> None:
        """
        :param subscriber: Subscriber to the distributed instance.
        """
        with self._lock_local:
            if subscriber not in self._subscribers:
                self._subscribers.append(subscriber)

    def _update(self) -> None:
        """
        Processes data in the incoming queue and passes them to subscribers.
        """
        # Take incoming data and pass to decider
        while not self._queue_incoming.empty():
            incoming: Dict[str, List[BoboRunSerial]] = \
                self._queue_incoming.get_nowait()

            completed: List[BoboRunSerial] = incoming[_KEY_COMPLETED]
            halted: List[BoboRunSerial] = incoming[_KEY_HALTED]
            updated: List[BoboRunSerial] = incoming[_KEY_UPDATED]

            logging.debug("{} Sending incoming queue data to subscribers"
                          .format(self._urn))

            if (
                    len(completed) > 0 or
                    len(halted) > 0 or
                    len(updated) > 0
            ):
                for subscriber in self._subscribers:
                    subscriber.on_distributed_update(
                        completed=completed,
                        halted=halted,
                        updated=updated)

    def on_decider_update(
            self,
            completed: List[BoboRunSerial],
            halted: List[BoboRunSerial],
            updated: List[BoboRunSerial],
            local: bool
    ) -> None:
        """
        :param completed: Locally completed runs.
        :param halted: Locally halted runs.
        :param updated: Locally updated runs.
        :param local: `True` if the Decider update occurred locally;
            `False` if the update occurred on a remote (distributed) instance.
        """
        with self._lock_local:
            if self._closed:
                raise BoboDistributedError(_EXC_CLOSED)

            if not self._running:
                raise BoboDistributedError(_EXC_NOT_RUNNING)

            # Prevents Decider from passing distributed data back
            if not local:
                return

            logging.debug("{} Adding local Decider changes "
                          "to outgoing queue".format(self._urn))

            # Take changes to decider and pass to outgoing

            outgoing: Dict[str, List[BoboRunSerial]] = {
                _KEY_COMPLETED: completed,
                _KEY_HALTED: halted,
                _KEY_UPDATED: updated
            }

            if not self._queue_outgoing.full():
                self._queue_outgoing.put_nowait(outgoing)

            else:
                errmsg = "Outgoing queue is full."
                logging.critical(errmsg)
                raise BoboDistributedSystemError(errmsg)

    @staticmethod
    def _now() -> int:
        """
        :return: The current time in seconds since epoch.
        """
        return int(time.time())

    def _tcp_outgoing(self):
        """
        Executed within a thread. It sends internal Decider state updates,
        stored in the outgoing queue, to external `BoboCEP` instances.
        """

        while True:
            with self._lock_in_out:
                # Thread has been marked to close
                if self._thread_closed:
                    return

                outlist: List[Tuple[BoboDeviceManager, int]] = []
                now: int = self._now()

                # Determine what to send to each device...
                for d in self._devices.values():
                    # Ignore self...
                    if d.urn == self._urn:
                        continue

                    comms_range: int = (now - d.last_comms)
                    attempt_range: int = (now - d.last_attempt)
                    queue_empty: bool = self._queue_outgoing.empty()

                    # If device is within the "RESYNC Period"...
                    if comms_range >= self._period_resync:
                        # ...and due to attempt another RESYNC...
                        if attempt_range >= self._attempt_resync:
                            outlist.append((d, _TYPE_RESYNC))

                    # If device is within the "PING Period"
                    # and there is nothing to SYNC...
                    elif comms_range >= self._period_ping and (
                            queue_empty and
                            d.size_stash() == 0
                    ):
                        # ...and due to attempt another PING...
                        if attempt_range >= self._attempt_ping:
                            outlist.append((d, _TYPE_PING))

                    # If device is within the "SYNC Period"
                    # or "PING Period" with something left to SYNC...
                    else:
                        # ...and there is data in the queue or
                        # data in the device's stash that is due to send...
                        if (not queue_empty) or (
                                d.size_stash() > 0 and
                                attempt_range >= self._attempt_stash
                        ):
                            outlist.append((d, _TYPE_SYNC))

            # Compiled SYNC data for sending to all devices
            cache_sync: Optional[Dict[str, List[BoboRunSerial]]] = None

            for d, msg_type in outlist:
                # Set flags
                msg_flags: int = 0

                if d.flag_reset:
                    msg_flags += _FLAG_RESET

                if msg_type == _TYPE_RESYNC:
                    # Stash contents not needed when in RESYNC Period
                    d.clear_stash()

                    # Collect and cache a snapshot of Decider
                    snapshot = self._decider.snapshot()
                    msg_json_resync = self._outgoing_to_json({
                        _KEY_COMPLETED: snapshot[0],
                        _KEY_HALTED: snapshot[1],
                        _KEY_UPDATED: snapshot[2]
                    })

                    # Send snapshot to remote
                    err: int = self._tcp_send(
                        d, msg_type, msg_flags, msg_json_resync)

                    # Update last comms on success
                    now = self._now()

                    if err == 0:
                        logging.debug("{} Resync SUCCESS: {}"
                                      .format(self._urn, d.urn))
                        d.last_comms = now

                        if (msg_flags & _FLAG_RESET) == _FLAG_RESET:
                            d.flag_reset = False

                    else:
                        logging.error(
                            "{} Resync FAILURE: {} (code {})"
                            .format(self._urn, d.urn, err))

                    d.last_attempt = now

                elif msg_type == _TYPE_PING:
                    # Send ping
                    err: int = self._tcp_send(d, msg_type, msg_flags, "{}")

                    # Update last comms on success
                    now = self._now()

                    if err == 0:
                        logging.debug("{} Ping SUCCESS: {}"
                                      .format(self._urn, d.urn))
                        d.last_comms = now

                        if (msg_flags & _FLAG_RESET) == _FLAG_RESET:
                            d.flag_reset = False

                    else:
                        logging.error(
                            "{} Ping FAILURE: {} (code {})"
                            .format(self._urn, d.urn, err))

                    d.last_attempt = now

                elif msg_type == _TYPE_SYNC:
                    # Get SYNC data to send (cached for all devices to use)
                    if cache_sync is None:
                        if not self._queue_outgoing.empty():
                            cache_sync = self._queue_outgoing.get_nowait()
                        else:
                            cache_sync = {
                                _KEY_COMPLETED: [],
                                _KEY_HALTED: [],
                                _KEY_UPDATED: []
                            }

                    # Get device's stash of unsent data
                    stash_c, stash_h, stash_u = d.stash()

                    # Add stash to new data to send
                    send_sync: Dict[str, List[BoboRunSerial]] = {
                        _KEY_COMPLETED:
                            cache_sync[_KEY_COMPLETED] + stash_c,
                        _KEY_HALTED:
                            cache_sync[_KEY_HALTED] + stash_h,
                        _KEY_UPDATED:
                            cache_sync[_KEY_UPDATED] + stash_u
                    }

                    # Send JSON sync
                    msg_json_sync: str = self._outgoing_to_json(send_sync)
                    err: int = self._tcp_send(
                        d, msg_type, msg_flags, msg_json_sync)

                    now = self._now()
                    if err == 0:
                        logging.debug("{} Sync SUCCESS: {}"
                                      .format(self._urn, d.urn))

                        # Update last comms on success and clear stash
                        d.clear_stash()
                        d.last_comms = now

                        if (msg_flags & _FLAG_RESET) == _FLAG_RESET:
                            d.flag_reset = False
                    else:
                        logging.error("{} Sync FAILURE: {} (code {})"
                                      .format(self._urn, d.urn, err))

                        # Append stash with cache_sync on error
                        d.append_stash(
                            completed=cache_sync[_KEY_COMPLETED],
                            halted=cache_sync[_KEY_HALTED],
                            updated=cache_sync[_KEY_UPDATED])

                    d.last_attempt = now

    def _tcp_send(self, d: BoboDeviceManager, msg_type: int,
                  msg_flags: int, msg_str: str) -> int:
        """
        Sends data via TCP.

        :param d: The device to send data to.
        :param msg_type: Message type.
        :param msg_flags: Message flags.
        :param msg_str: The data to send.
        :return: 0 for success; 1 for timeout error; 2 for system error.
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        try:
            logging.debug("{} Sending to: {} ({}:{})"
                          .format(self._urn, d.urn, d.addr, d.port))

            s.settimeout(self._timeout_connect)
            s.connect((d.addr, d.port))
            s.settimeout(None)

            mydev = self._devices[self._urn]

            msg_str = "{} {} {} {} {}".format(
                mydev.urn,
                mydev.id_key,
                msg_type,
                msg_flags,
                msg_str)

            msg_bytes = self._crypto.encrypt(msg_str)

            s.settimeout(self._timeout_send)
            s.sendall(msg_bytes)
            s.settimeout(None)

            return 0

        except TimeoutError:
            # Raised when a system function timed out at the system level.
            return 1

        except OSError:
            # Raised when a system function returns a system-related error.
            return 2

        finally:
            s.close()

    def _tcp_incoming(self) -> None:
        """
        Executed within a thread. It accepts outside connections and stores
        their data in the incoming queue, if valid.
        """

        mydev = self._devices[self._urn]

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind((mydev.addr, mydev.port))

        # queue requests before refusing outside connections
        s.listen(self._max_listen)

        try:
            while True:
                with self._lock_in_out:
                    if self._thread_closed:
                        return

                try:
                    s.settimeout(self._timeout_accept)

                    client_s, client_origin = s.accept()
                    client_addr = client_origin[0]
                    client_port = client_origin[1]

                    s.settimeout(None)
                    client_accepted = int(time.time())

                    logging.debug("{} Incoming client: {}:{}"
                                  .format(self._urn, client_addr, client_port))

                    self._tcp_incoming_handle_client(
                        client_s, client_addr, client_accepted)

                except (BoboDistributedSystemError,
                        BoboDistributedTimeoutError) as e:
                    # From _tcp_incoming_handle_client
                    logging.error("{} {}: {}"
                                  .format(self._urn, e.__class__.__name__, e))

                except socket.timeout:
                    # From s.accept()
                    # Timeout is added here so that the thread can react to
                    # _thread_closed=True when no data received for a while
                    pass

                except Exception as e:
                    # From other
                    logging.error("{} {}: {}"
                                  .format(self._urn, e.__class__.__name__, e))
        finally:
            s.close()

    def _tcp_incoming_handle_client(
            self, client_s, client_addr, client_accepted: int) -> None:
        """
        Handles an incoming client request.
        :param client_s: The client socket.
        :param client_addr: The client address.
        :param client_accepted: The time when the client request was accepted.
        """
        try:
            all_bytes = bytearray()

            while True:
                now = int(time.time())
                elapse = (now - client_accepted)

                if elapse >= self._timeout_receive:
                    raise BoboDistributedTimeoutError(
                        "Message timeout ({} seconds)"
                        .format(elapse))

                bytes_msg = client_s.recv(self._recv_bytes)
                all_bytes.extend(bytes_msg)

                # If bytes received so far are at least the minimum length
                # and end with the expected end bytes.
                if (
                        len(bytes_msg) >= self._crypto.min_length() and
                        bytes_msg[-len(self._crypto.end_bytes()):]
                        == self._crypto.end_bytes()
                ):

                    try:
                        plaintext = self._crypto.decrypt(all_bytes)
                    except ValueError as e:
                        raise BoboDistributedSystemError(
                            "Failed to unwrap incoming message (bytes: {})"
                            .format(e))

                    pt_urn, pt_id, pt_type, pt_flags, pt_json = \
                        self._split_plaintext(plaintext)

                    logging.debug(
                        "{} Incoming: urn={}, type={}, flags={}"
                        .format(self._urn, pt_urn, pt_type, pt_flags))

                    # Check if URN is NOT a recognised device
                    if pt_urn not in self._devices:
                        raise BoboDistributedSystemError(
                            "Unknown device URN '{}'.".format(pt_urn))

                    device: BoboDeviceManager = self._devices[pt_urn]

                    # Check if ID key DOES NOT MATCH expected key for URN
                    if pt_id != device.id_key:
                        raise BoboDistributedSystemError(
                            "Invalid ID key for URN '{}'".format(device.urn))

                    # Update address if remote address has changed
                    if client_addr != device.addr:
                        logging.debug(
                            "{} Device {} addr update: from '{}' to '{}'"
                            .format(self._urn,
                                    device.urn, device.addr, client_addr))
                        device.addr = client_addr

                    if pt_type == _TYPE_SYNC or pt_type == _TYPE_RESYNC:
                        incoming = self._incoming_from_json(pt_json)

                        logging.debug("{} Data from {}: {}"
                                      .format(self._urn, pt_urn, incoming))

                        # Add incoming data to queue
                        if not self._queue_incoming.full():
                            self._queue_incoming.put_nowait(incoming)
                        else:
                            errmsg = "Incoming queue is full."
                            logging.critical(errmsg)
                            raise BoboDistributedSystemError(errmsg)

                    # (Nothing to do on PING; it exists for outgoing.)
                    if pt_type == _TYPE_PING:
                        logging.debug("{} Pinged by: {}"
                                      .format(self._urn, pt_urn))

                    if (pt_flags & _FLAG_RESET) == _FLAG_RESET:
                        # Clears comms for device to trigger RESYNC
                        logging.debug("{} Reset for: {}"
                                      .format(self._urn, pt_urn))
                        device.clear_last()

                    break
        finally:
            client_s.close()

    def _split_plaintext(self, plaintext: str) \
            -> Tuple[str, str, int, int, str]:
        """
        :param plaintext: The string to split.
        :return: Tuple containing: URN, ID, type, flags, and JSON message.
        """
        ix_delim = []

        for i, c in enumerate(plaintext):
            if c == ' ':
                ix_delim.append(i)

            if len(ix_delim) == 4:
                break

        if len(ix_delim) != 4:
            raise BoboDistributedError(
                "Invalid plaintext message: {}".format(plaintext))

        # pt_urn, pt_id,, pt_type, pt_flags, pt_json
        return (
            plaintext[:ix_delim[0]],
            plaintext[ix_delim[0] + 1:ix_delim[1]],
            int(plaintext[ix_delim[1] + 1:ix_delim[2]]),
            int(plaintext[ix_delim[2] + 1:ix_delim[3]]),
            plaintext[ix_delim[3] + 1:]
        )

    def _incoming_from_json(
            self, msg_str: str) -> Dict[str, List[BoboRunSerial]]:
        """
        :param msg_str: Incoming JSON string.
        :return: Incoming Decider update information.
        """
        try:
            return json.loads(msg_str, cls=_IncomingJSONDecoder)

        except (BoboJSONableError, TypeError) as e:
            raise BoboDistributedSystemError(
                "Failed to parse incoming JSON: {}".format(e))

    def _outgoing_to_json(
            self, msg: Dict[str, List[BoboRunSerial]]) -> str:
        """
        :param msg: Outgoing Decider update information.
        :return: Outgoing JSON string.
        """
        try:
            return json.dumps(msg, cls=_OutgoingJSONEncoder)

        except (BoboJSONableError, TypeError) as e:
            raise BoboDistributedSystemError(
                "Failed to generate outgoing JSON: {}".format(e))

    def join(self) -> None:
        """
        Joins with the incoming and outgoing threads.
        """
        with self._lock_local:
            self._thread_incoming.join()
            self._thread_outgoing.join()

    def close(self) -> None:
        """
        Closes the distributed instance.
        """
        with self._lock_local:
            if self._closed:
                return

            try:
                logging.debug("{} Close distributed".format(self._urn))

                with self._lock_in_out:
                    self._thread_closed = True

            finally:
                self._closed = True

    def is_closed(self) -> bool:
        """
        :return: `True` if distributed is closed; `False` otherwise.
        """
        with self._lock_local:
            return self._closed

    def size_incoming(self) -> int:
        """
        :return: Size of incoming queue.
        """
        with self._lock_local:
            return self._queue_incoming.qsize()

    def size_outgoing(self) -> int:
        """
        :return: Size of outgoing queue
        """
        with self._lock_local:
            return self._queue_outgoing.qsize()



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Engine task that detects patterns in data and decides whether a complex event
has manifest.
"""

from collections import deque
from queue import Queue
from threading import RLock
from typing import Tuple, Dict, List, Optional, Deque

from bobocep.cep.engine.decider.pubsub import BoboDeciderPublisher, \
    BoboDeciderSubscriber
from bobocep.cep.engine.decider.run import BoboRun
from bobocep.cep.engine.decider.runserial import BoboRunSerial
from bobocep.cep.engine.receiver.pubsub import BoboReceiverSubscriber
from bobocep.cep.engine.task import BoboEngineTaskError, BoboEngineTask
from bobocep.cep.event import BoboHistory, BoboEvent
from bobocep.cep.gen.event_id import BoboGenEventID
from bobocep.cep.phenom.pattern.pattern import BoboPattern
from bobocep.cep.phenom.phenom import BoboPhenomenon
from bobocep.dist.pubsub import BoboDistributedSubscriber

_EXC_PHENOM_NAME_DUP = "duplicate name in phenomena: {}"
_EXC_QUEUE_FULL = "queue is full (max size: {})"
_EXC_RUN_NOT_FOUND = "run {} not found for phenomenon {}, pattern {}"
_EXC_RUN_EXISTS = "run {} already exists for phenomenon {}, pattern {}"


class BoboDeciderError(BoboEngineTaskError):
    """
    A decider task error.
    """


class BoboDecider(BoboEngineTask,
                  BoboDeciderPublisher,
                  BoboReceiverSubscriber,
                  BoboDistributedSubscriber):
    """
    A decider task.
    """

    def __init__(self,
                 phenomena: List[BoboPhenomenon],
                 gen_event_id: BoboGenEventID,
                 gen_run_id: BoboGenEventID,
                 max_cache: int = 0,
                 max_size: int = 0):
        """
        :param phenomena: List of phenomena.
        :param gen_event_id: Event ID generator.
        :param gen_run_id: Run ID generator.
        :param max_cache: Max cache size (<=0 means no caching).
            Default: 0.
        :param max_size: Max queue size.
            Default: 0 (unbounded).
        """
        super().__init__()

        self._lock: RLock = RLock()
        self._closed: bool = False
        self._subscribers: List[BoboDeciderSubscriber] = []

        self._phenomena: Dict[str, BoboPhenomenon] = {}

        for phenom in phenomena:
            if phenom.name not in self._phenomena:
                self._phenomena[phenom.name] = phenom
            else:
                raise BoboDeciderError(
                    _EXC_PHENOM_NAME_DUP.format(phenom.name))

        self._gen_event_id: BoboGenEventID = gen_event_id
        self._gen_run_id: BoboGenEventID = gen_run_id
        # Phenomenon Name => Pattern Name => Run ID => Run
        self._runs: Dict[str, Dict[str, Dict[str, BoboRun]]] = {}
        self._stub_history: BoboHistory = BoboHistory({})
        self._max_size: int = max(0, max_size)
        self._queue: Queue[BoboEvent] = Queue(self._max_size)

        self._caching: bool = max_cache > 0
        self._cache_completed: Optional[Deque[BoboRunSerial]] = \
            deque(maxlen=max_cache) if self._caching else None
        self._cache_halted: Optional[Deque[BoboRunSerial]] = \
            deque(maxlen=max_cache) if self._caching else None

    def subscribe(self, subscriber: BoboDeciderSubscriber) -> None:
        """
        :param subscriber: Subscriber to Decider data.
        """
        with self._lock:
            if subscriber not in self._subscribers:
                self._subscribers.append(subscriber)

    def update(self) -> bool:
        """
        Performs an update cycle of the decider that takes an event from its
        queue and checks it against phenomena and existing runs.

        :return: True if an internal state change occurred during the update;
            False otherwise.
        """
        with self._lock:
            if self._closed:
                return False

            if not self._queue.empty():
                # Process event and collect changes to decider
                rl_completed, rl_halted, rl_updated = \
                    self._process_event(self._queue.get_nowait())

                completed: List[BoboRunSerial] = \
                    [run_c.serialize() for run_c in rl_completed]
                halted: List[BoboRunSerial] = \
                    [run_h.serialize() for run_h in rl_halted]
                updated: List[BoboRunSerial] = \
                    [run_u.serialize() for run_u in rl_updated]

                # Cache local changes
                self._maybe_cache(completed, halted)

                # Only notify subscribers if at least one list has values
                internal_state_change = any(len(rl) > 0 for rl in
                                            [completed, halted, updated])
                if internal_state_change:
                    for subscriber in self._subscribers:
                        subscriber.on_decider_update(
                            completed=completed,
                            halted=halted,
                            updated=updated,
                            local=True)

                return internal_state_change
            return False

    def snapshot(self) -> Tuple[
        List[BoboRunSerial], List[BoboRunSerial], List[BoboRunSerial]
    ]:
        """
        A snapshot of the current state of the Decider.

        :return: Tuple of cached completed, cached halted, and
            currently partially-completed runs.
            If caching is disabled, the first two lists will be empty.
        """
        with self._lock:
            if self._closed:
                return [], [], []

            if (
                    self._caching and
                    self._cache_completed is not None and
                    self._cache_halted is not None
            ):
                # Get completed from cache
                r_completed = [c for c in self._cache_completed] \
                    if self._caching else []

                # Get halted from cache
                r_halted = [h for h in self._cache_halted] \
                    if self._caching else []
            else:
                r_completed = []
                r_halted = []

            # Get updated from the current state of partially-completed runs
            r_updated = []
            for k_phenom in self._runs.keys():
                for k_pattern in self._runs[k_phenom].keys():
                    for k_id in self._runs[k_phenom][k_pattern].keys():
                        r_updated.append(
                            self._runs[k_phenom][k_pattern][k_id].serialize())

            return r_completed, r_halted, r_updated

    def _maybe_cache(
            self,
            completed: List[BoboRunSerial],
            halted: List[BoboRunSerial]) -> None:
        """
        Caches completed and halted runs, if caching is enabled.#

        :param completed: Completed runs.
        :param halted: Halted runs.
        """
        if (
                self._caching and
                self._cache_completed is not None and
                self._cache_halted is not None
        ):
            # Cache runs that have been locally completed
            for c in completed:
                self._cache_completed.append(c)

            # Cache runs that have been locally halted
            for h in halted:
                self._cache_halted.append(h)

    def on_receiver_update(self, event: BoboEvent) -> None:
        """
        :param event: Event from Receiver.
        """
        with self._lock:
            if self._closed:
                return

            if not self._queue.full():
                self._queue.put(event)
            else:
                raise BoboDeciderError(
                    _EXC_QUEUE_FULL.format(self._max_size))

    def _maybe_check_against_cache(
            self,
            completed: List[BoboRunSerial],
            halted: List[BoboRunSerial],
            updated: List[BoboRunSerial]) -> Tuple[List[BoboRunSerial],
                                                   List[BoboRunSerial],
                                                   List[BoboRunSerial]]:
        """
        Compares run changes that occurred remotely with local run states.

        :param completed: Completed runs.
        :param halted: Halted runs.
        :param updated: Updated runs.

        :return: The original lists but with the following changes:
            (1) completed runs kept if they have not been complete locally;
            (2) halted runs kept if not halted locally; and
            (3) updated runs kept if not completed or halted locally.
        """
        if (
                self._caching and
                self._cache_completed is not None and
                self._cache_halted is not None
        ):
            # Keep completed IDs if not completed locally
            # Complete takes precedent over halt and update
            completed = [
                comp for comp in completed
                if (
                    not any(comp.run_id == cache_comp.run_id
                            for cache_comp in self._cache_completed)
                )]

            # Keep halted IDs if not completed and not halted locally
            # Halt takes precedent over update
            halted = [ch for ch in halted
                      if ch not in self._cache_completed and
                      ch not in self._cache_halted]

            # Keep updated IDs if not completed and not halted locally
            updated = [cu for cu in updated
                       if cu not in self._cache_completed and
                       cu not in self._cache_halted]

        return completed, halted, updated

    def on_distributed_update(
            self,
            completed: List[BoboRunSerial],
            halted: List[BoboRunSerial],
            updated: List[BoboRunSerial]) -> None:
        """
        :param completed: Completed runs.
        :param halted: Halted runs.
        :param updated: Updated runs.
        """

        with self._lock:
            if self._closed:
                return

            # Remove any invalid remote changes
            completed, halted, updated = \
                self._maybe_check_against_cache(completed, halted, updated)

            # Remove runs that were completed remotely
            for rc in completed:
                self._remove_run(
                    rc.phenomenon_name,
                    rc.pattern_name,
                    rc.run_id,
                    quiet=True)

            # Remove runs that were halted remotely
            for rh in halted:
                self._remove_run(
                    rh.phenomenon_name,
                    rh.pattern_name,
                    rh.run_id,
                    quiet=True)

            # Update existing runs, or add new run that was started remotely
            update_remove_indices = []

            for i, runtup in enumerate(updated):
                urun: Optional[BoboRun] = self.run_at(
                    runtup.phenomenon_name,
                    runtup.pattern_name,
                    runtup.run_id)

                if urun is not None:
                    # Update existing run
                    # Push local run forward to index and history from remote
                    if runtup.block_index > urun.block_index:
                        urun.set_block(
                            block_index=runtup.block_index,
                            history=runtup.history)

                else:
                    pattern: Optional[BoboPattern] = self._get_pattern(
                        runtup.phenomenon_name, runtup.pattern_name)

                    # Ignore if pattern does not exist - it may have been
                    # removed from the decider
                    if pattern is None:
                        update_remove_indices.append(i)
                        continue

                    # Add new run that was started remotely
                    newrun = BoboRun(
                        run_id=runtup.run_id,
                        phenomenon_name=runtup.phenomenon_name,
                        pattern=pattern,
                        block_index=runtup.block_index,
                        history=runtup.history)

                    self._add_run(runtup.phenomenon_name, runtup.pattern_name,
                                  newrun)

            # Remove any updates for which a pattern could not be found
            for i in sorted(update_remove_indices, reverse=True):
                del updated[i]

            # Cache remote changes
            self._maybe_cache(completed, halted)

            # Notify subscribers
            for subscriber in self._subscribers:
                subscriber.on_decider_update(
                    completed=completed,
                    halted=halted,
                    updated=updated,
                    local=False)

    def _get_pattern(self,
                     phenomenon_name: str,
                     pattern_name: str) -> Optional[BoboPattern]:
        """
        :param phenomenon_name: A phenomenon name.
        :param pattern_name: A pattern name.
        :return: A BoboPattern instance corresponding to the
            phenomenon and pattern name.
        """
        if phenomenon_name in self._phenomena:
            for pattern in self._phenomena[phenomenon_name].patterns:
                if pattern.name == pattern_name:
                    return pattern
        return None

    def phenomena(self) -> Tuple[BoboPhenomenon, ...]:
        """
        :return: All phenomena under consideration by the decider.
        """
        with self._lock:
            return tuple(self._phenomena.values())

    def all_runs(self) -> Tuple[BoboRun, ...]:
        """
        :return: All active runs in the decider.
        """
        with self._lock:
            runs: List[BoboRun] = []
            for phenomenon_name, dict_patterns in self._runs.items():
                for pattern_name, dict_runs in dict_patterns.items():
                    for _, drun in dict_runs.items():
                        runs.append(drun)
            return tuple(runs)

    def runs_from(self,
                  phenomenon_name: str,
                  pattern_name: str) -> Tuple[BoboRun, ...]:
        """
        :param phenomenon_name: A phenomenon name.
        :param pattern_name: A pattern name.
        :return: The runs associated with the given
            phenomenon and pattern name.
        """
        with self._lock:
            if (
                    phenomenon_name in self._runs and
                    pattern_name in self._runs[phenomenon_name]
            ):
                return tuple(
                    self._runs[phenomenon_name][pattern_name].values())
            return tuple()

    def run_at(self,
               phenomenon_name: str,
               pattern_name: str,
               run_id: str) -> Optional[BoboRun]:
        """
        :param phenomenon_name: A phenomenon name.
        :param pattern_name: A pattern name.
        :param run_id: A run ID.
        :return: A run associated with the given phenomenon and pattern name;
            or None if no such run exists.
        """
        with self._lock:
            if (
                    phenomenon_name in self._runs and
                    pattern_name in self._runs[phenomenon_name] and
                    run_id in self._runs[phenomenon_name][pattern_name]
            ):
                return self._runs[phenomenon_name][pattern_name][run_id]
            return None

    def size(self) -> int:
        """
        :return: The total number of events in the decider's queue.
        """
        with self._lock:
            return self._queue.qsize()

    def close(self) -> None:
        """
        Closes the Decider.
        """
        with self._lock:
            self._closed = True

    def is_closed(self) -> bool:
        """
        :return: `True` if decider is set to close; `False` otherwise.
        """
        with self._lock:
            return self._closed

    def _process_event(self, event: BoboEvent) -> \
            Tuple[List[BoboRun], List[BoboRun], List[BoboRun]]:
        """
        :param event: An event.

        :return: Runs that had a state change due to the event.
        """
        r_halt_com, r_halt_incom, r_upd = self._check_against_runs(event)
        p_halt_com, p_upd = self._check_against_patterns(event)

        return (r_halt_com + p_halt_com), r_halt_incom, (r_upd + p_upd)

    def _check_against_runs(self, event: BoboEvent) -> \
            Tuple[List[BoboRun], List[BoboRun], List[BoboRun]]:
        """
        :param event: An event.

        :return: Runs that had a state change due to the event.
        """
        runs_halted_complete: List[BoboRun] = []
        runs_halted_incomplete: List[BoboRun] = []
        runs_updated: List[BoboRun] = []

        runs_to_remove: List[Tuple[str, str, str]] = []

        for phenomenon_name, dict_patterns in self._runs.items():
            for pattern_name, dict_runs in dict_patterns.items():
                for _, run in dict_runs.items():
                    # If an internal state change occurs in the run...
                    run_eval: bool
                    try:
                        run_eval = run.process(event)
                    except (Exception,):
                        continue

                    # ...determine if completed, halted, or updated.
                    if run_eval:
                        if run.is_halted():
                            runs_to_remove.append((
                                phenomenon_name, pattern_name, run.run_id))
                            if run.is_complete():
                                runs_halted_complete.append(run)
                            else:
                                runs_halted_incomplete.append(run)
                        else:
                            runs_updated.append(run)

        for phenomenon_name, pattern_name, run_id in runs_to_remove:
            self._remove_run(phenomenon_name, pattern_name, run_id)

        return runs_halted_complete, runs_halted_incomplete, runs_updated

    def _check_against_patterns(self, event: BoboEvent) -> \
            Tuple[List[BoboRun], List[BoboRun]]:
        """
        :param event: An event.

        :return: Newly-created runs due to the event.
        """
        runs_halted_complete: List[BoboRun] = []
        runs_updated: List[BoboRun] = []

        for phenomenon in self._phenomena.values():
            for pattern in phenomenon.patterns:
                # If any predicate in a block evaluates to True...
                any_eval: bool = False
                for predicate in pattern.blocks[0].predicates:
                    try:
                        if predicate.evaluate(event, self._stub_history):
                            any_eval = True
                            break
                    except (Exception,):
                        pass

                # ...create a run.
                if any_eval:
                    newrun = BoboRun(
                        run_id=self._gen_run_id.generate(),
                        phenomenon_name=phenomenon.name,
                        pattern=pattern,
                        block_index=1,
                        history=BoboHistory({
                            pattern.blocks[0].group: [event]
                        }))

                    if newrun.is_halted() and newrun.is_complete():
                        runs_halted_complete.append(newrun)
                    else:
                        self._add_run(phenomenon.name, pattern.name, newrun)
                        runs_updated.append(newrun)

        return runs_halted_complete, runs_updated

    def _add_run(self,
                 phenomenon_name: str,
                 pattern_name: str,
                 newrun: BoboRun) -> None:
        """
        Adds new run to Decider.

        :param phenomenon_name: Phenomenon associated with new run.
        :param pattern_name: Pattern associated with new run.
        :param newrun: Run to add.
        """
        if phenomenon_name not in self._runs:
            self._runs[phenomenon_name] = {}

        if pattern_name not in self._runs[phenomenon_name]:
            self._runs[phenomenon_name][pattern_name] = {}

        if newrun.run_id not in self._runs[phenomenon_name][pattern_name]:
            self._runs[phenomenon_name][pattern_name][newrun.run_id] = newrun
        else:
            raise BoboDeciderError(_EXC_RUN_EXISTS.format(
                newrun.run_id, phenomenon_name, pattern_name))

    def _remove_run(self,
                    phenomenon_name: str,
                    pattern_name: str,
                    run_id: str,
                    quiet: bool = False) -> None:
        """
        :param phenomenon_name: The phenomenon name.
        :param pattern_name: The pattern name.
        :param run_id: The run ID.
        :param quiet: If True, do not raise exceptions; False to raise them.

        :raises BoboDeciderError: Run is not found.
        """

        if (
                phenomenon_name in self._runs and
                pattern_name in self._runs[phenomenon_name] and
                run_id in self._runs[phenomenon_name][pattern_name]
        ):
            del self._runs[phenomenon_name][pattern_name][run_id]
        else:
            if not quiet:
                raise BoboDeciderError(_EXC_RUN_NOT_FOUND.format(
                    run_id, phenomenon_name, pattern_name))



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Crypto imports.
"""

from bobocep.dist.crypto.aes import BoboDistributedCryptoAES
from bobocep.dist.crypto.crypto import BoboDistributedCrypto, \
    BoboDistributedCryptoError



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Simple setup.
"""

from typing import List, Optional, Tuple

from bobocep.cep.action.handler import BoboActionHandler
from bobocep.cep.engine.decider.decider import BoboDecider
from bobocep.cep.engine.engine import BoboEngine
from bobocep.cep.engine.forwarder.forwarder import BoboForwarder
from bobocep.cep.engine.producer.producer import BoboProducer
from bobocep.cep.engine.receiver.receiver import BoboReceiver
from bobocep.cep.engine.receiver.validator import BoboValidator, \
    BoboValidatorAll, BoboValidatorJSONable
from bobocep.cep.gen.event import BoboGenEvent
from bobocep.cep.gen.event_id import BoboGenEventIDUnique
from bobocep.cep.gen.timestamp import BoboGenTimestampEpoch
from bobocep.cep.phenom.phenom import BoboPhenomenon
from bobocep.dist.crypto.aes import BoboDistributedCryptoAES
from bobocep.dist.device import BoboDevice
from bobocep.dist.tcp import BoboDistributedTCP
from bobocep.setup.setup import BoboSetup


class BoboSetupSimple(BoboSetup):
    """A simple setup to make configuration easier."""

    def __init__(
            self,
            phenomena: List[BoboPhenomenon],
            handler: BoboActionHandler,
            validator: Optional[BoboValidator] = None,
            gen_event: Optional[BoboGenEvent] = None,
            urn: Optional[str] = None
    ):
        """
        :param phenomena: A list of phenomena.
        :param handler: An action handler.
        :param validator: A data validator for the engine's Receiver task.
            Default: BoboValidatorAll.
        :param gen_event: An event generator.
            Default: None.
        :param urn: A URN for ID generation.
        """
        super().__init__()

        self._phenomena: List[BoboPhenomenon] = phenomena
        self._validator: BoboValidator = validator \
            if validator is not None else BoboValidatorAll()
        self._handler: BoboActionHandler = handler
        self._gen_event: Optional[BoboGenEvent] = gen_event
        self._urn: Optional[str] = urn

    def generate(self) -> BoboEngine:
        """
        :return: The CEP engine.
        """
        gen_event_id = BoboGenEventIDUnique(self._urn)
        gen_run_id = BoboGenEventIDUnique(self._urn)
        gen_timestamp = BoboGenTimestampEpoch()

        receiver = BoboReceiver(
            validator=self._validator,
            gen_event_id=gen_event_id,
            gen_timestamp=gen_timestamp,
            gen_event=self._gen_event)

        decider = BoboDecider(
            phenomena=self._phenomena,
            gen_event_id=gen_event_id,
            gen_run_id=gen_run_id)

        producer = BoboProducer(
            phenomena=self._phenomena,
            gen_event_id=gen_event_id,
            gen_timestamp=gen_timestamp)

        forwarder = BoboForwarder(
            phenomena=self._phenomena,
            handler=self._handler,
            gen_event_id=gen_event_id,
            gen_timestamp=gen_timestamp)

        engine = BoboEngine(
            receiver=receiver,
            decider=decider,
            producer=producer,
            forwarder=forwarder)

        return engine


class BoboSetupSimpleDistributed(BoboSetup):
    """
    A simple setup to make distributed configuration easier.
    """

    def __init__(
            self,
            phenomena: List[BoboPhenomenon],
            handler: BoboActionHandler,
            urn: str,
            devices: List[BoboDevice],
            aes_key: str,
            validator: Optional[BoboValidatorJSONable] = None,
            gen_event: Optional[BoboGenEvent] = None):
        """
        :param phenomena: A list of phenomena.
        :param handler: An action handler.
        :param urn: A URN that is unique across devices in the network.
        :param devices: Devices in the network (including this device).
        :param aes_key: The AES key to use for encryption.
        :param validator: A data validator for the engine's Receiver task.
            Default: BoboValidatorAll.
        :param gen_event: An event generator.
            Default: None.
        """
        super().__init__()

        self._setup_simple = BoboSetupSimple(
            phenomena=phenomena,
            handler=handler,
            validator=validator if validator is not None else
            BoboValidatorJSONable(),
            gen_event=gen_event,
            urn=urn
        )

        self._urn: str = urn
        self._devices: List[BoboDevice] = devices
        self._aes_key: str = aes_key

    def generate(self) -> Tuple[BoboEngine, BoboDistributedTCP]:
        """
        :return: The CEP engine and distributed TCP instance
            with AES encryption.
        """
        engine: BoboEngine = self._setup_simple.generate()

        distributed: BoboDistributedTCP = BoboDistributedTCP(
            urn=self._urn,
            decider=engine.decider,
            devices=self._devices,
            crypto=BoboDistributedCryptoAES(aes_key=self._aes_key))

        engine.decider.subscribe(distributed)
        distributed.subscribe(engine.decider)

        return engine, distributed



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Device manager.
"""

from threading import RLock
from typing import List, Tuple

from bobocep.cep.engine.decider.runserial import BoboRunSerial
from bobocep.dist.device import BoboDevice


class BoboDeviceManager:
    """
    Manages information about a BoboCEP device on the network.
    """

    def __init__(self,
                 device: BoboDevice,
                 flag_reset: bool):
        """
        :param device: The device to manage.
        :param flag_reset: If `True`, it indicates to the BoboDistributed
            instance that the next message to the device should set a flag to
            reset its data on the sending device;
            'False' indicates that no flag should be set.
        """
        super().__init__()
        self._lock: RLock = RLock()

        self._device: BoboDevice = device
        self._flag_reset: bool = flag_reset

        self._last_comms: int = 0
        self._last_attempt: int = 0

        self._stash_completed: List[BoboRunSerial] = []
        self._stash_halted: List[BoboRunSerial] = []
        self._stash_updated: List[BoboRunSerial] = []

    @property
    def addr(self) -> str:
        """
        :return: Device address.
        """
        with self._lock:
            return self._device.addr

    @addr.setter
    def addr(self, addr: str) -> None:
        """
        Set device address.

        :param addr: The new address.
        """
        with self._lock:
            self._device.addr = addr

    @property
    def port(self) -> int:
        """
        :return: Device port.
        """
        return self._device.port

    @property
    def urn(self) -> str:
        """
        :return: Device URN.
        """
        return self._device.urn

    @property
    def id_key(self) -> str:
        """
        :return: Device ID key.
        """
        return self._device.id_key

    @property
    def flag_reset(self) -> bool:
        """
        :return: `True` if reset flag should be set; `False` otherwise.
        """
        with self._lock:
            return self._flag_reset

    @flag_reset.setter
    def flag_reset(self, flag_reset: bool) -> None:
        """
        :param flag_reset: `True` to set flag for resetting the device.
        """
        with self._lock:
            self._flag_reset = flag_reset

    @property
    def last_comms(self) -> int:
        """
        :return: Time of last communication with device.
        """
        with self._lock:
            return self._last_comms

    @last_comms.setter
    def last_comms(self, last_comms: int) -> None:
        """
        :param last_comms: The last communication with the device.
        """
        with self._lock:
            self._last_comms = max(0, last_comms)

    @property
    def last_attempt(self) -> int:
        """
        :return: Last attempted communication with the device.
        """
        with self._lock:
            return self._last_attempt

    @last_attempt.setter
    def last_attempt(self, last_attempt: int) -> None:
        """
        :return: Time of last communication attempt with device.
        """
        with self._lock:
            self._last_attempt = max(0, last_attempt)

    def clear_last(self) -> None:
        """
        Clears both the last communication and last attempted communication
        time by setting them both to 0.
        """
        with self._lock:
            self._last_comms = 0
            self._last_attempt = 0

    def stash(self) -> Tuple[
        List[BoboRunSerial], List[BoboRunSerial], List[BoboRunSerial]
    ]:
        """
        :return: The device's stash.
        """
        with self._lock:
            return self._stash_completed, \
                self._stash_halted, \
                self._stash_updated

    def append_stash(self,
                     completed: List[BoboRunSerial],
                     halted: List[BoboRunSerial],
                     updated: List[BoboRunSerial]) -> None:
        """
        Append runs to stash.

        :param completed: Completed runs to append.
        :param halted: Halted runs to append.
        :param updated: Updated runs to append.
        """
        with self._lock:
            self._stash_completed.extend(completed)
            self._stash_halted.extend(halted)
            self._stash_updated.extend(updated)

    def size_stash(self) -> int:
        """
        :return: The stash size, equal to all completed, halted,
            and updated runs in the stash.
        """
        with self._lock:
            return len(self._stash_completed) + \
                len(self._stash_halted) + \
                len(self._stash_updated)

    def clear_stash(self) -> None:
        """
        Removes all items in the stash.
        """
        with self._lock:
            self._stash_completed = []
            self._stash_halted = []
            self._stash_updated = []



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Simple event.
"""

from json import loads, dumps
from typing import Any

from bobocep.cep.event.event import BoboEvent


class BoboEventSimple(BoboEvent):
    """
    A simple event.
    """

    TYPE_SIMPLE = "type_simple"

    def __init__(self,
                 event_id: str,
                 timestamp: int,
                 data: Any):
        """
        :param event_id: The event ID.
        :param timestamp: The event timestamp.
        :param data: The event data.
        """
        super().__init__(
            event_id=event_id,
            timestamp=timestamp,
            data=data)

    def cast(self, dtype: type) -> 'BoboEventSimple':
        """
        :param dtype: The type to which the event's data is cast.
        :return: A new BoboEventSimple instance with its data cast to `dtype`
            and all other properties identical to the original event.
        """
        return BoboEventSimple(
            event_id=self._event_id,
            timestamp=self._timestamp,
            data=dtype(self._data)
        )

    def to_json_dict(self) -> dict:
        """
        :return: A JSON `dict` representation of the event.
        """
        return {
            self.EVENT_TYPE: self.TYPE_SIMPLE,
            self.EVENT_ID: self.event_id,
            self.TIMESTAMP: self.timestamp,
            self.DATA: self.data
        }

    def to_json_str(self) -> str:
        """
        :return: A JSON `str` representation of the event.
        """
        return dumps(self.to_json_dict(), default=lambda o: o.to_json_str())

    @staticmethod
    def from_json_str(j: str) -> 'BoboEventSimple':
        """
        :param j: A JSON `str` representation of the event.
        :return: A new instance of the event type.
        """
        return BoboEventSimple.from_json_dict(loads(j))

    @staticmethod
    def from_json_dict(d: dict) -> 'BoboEventSimple':
        """
        :param d: A JSON `dict` representation of the event.
        :return: A new instance of the event type.
        """
        return BoboEventSimple(
            event_id=d[BoboEventSimple.EVENT_ID],
            timestamp=d[BoboEventSimple.TIMESTAMP],
            data=d[BoboEventSimple.DATA]
        )

    def __str__(self) -> str:
        """
        :return: A JSON `str` representation of the event.
        """
        return self.to_json_str()



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Engine task that generates complex events and triggers actions.
"""

from queue import Queue
from threading import RLock
from typing import List, Dict, Tuple

from bobocep.cep.engine.decider.pubsub import BoboDeciderSubscriber
from bobocep.cep.engine.decider.runserial import BoboRunSerial
from bobocep.cep.engine.producer.pubsub import BoboProducerPublisher, \
    BoboProducerSubscriber
from bobocep.cep.engine.task import BoboEngineTaskError, BoboEngineTask
from bobocep.cep.event import BoboEventComplex
from bobocep.cep.gen.event_id import BoboGenEventID
from bobocep.cep.gen.timestamp import BoboGenTimestamp
from bobocep.cep.phenom.phenom import BoboPhenomenon

_EXC_PHENOM_NAME_DUP = "duplicate name in phenomena: {}"
_EXC_QUEUE_FULL = "queue is full (max size: {})"


class BoboProducerError(BoboEngineTaskError):
    """
    A producer task error.
    """


class BoboProducer(BoboEngineTask,
                   BoboProducerPublisher,
                   BoboDeciderSubscriber):
    """
    A producer task.
    """

    def __init__(self,
                 phenomena: List[BoboPhenomenon],
                 gen_event_id: BoboGenEventID,
                 gen_timestamp: BoboGenTimestamp,
                 max_size: int = 0):
        """
        :param phenomena: List of phenomena.
        :param gen_event_id: Event ID generator.
        :param gen_timestamp: Timestamp generator.
        :param max_size: Maximum queue size.
            Default: 0 (unbounded).
        """
        super().__init__()
        self._lock: RLock = RLock()
        self._closed: bool = False
        self._subscribers: List[BoboProducerSubscriber] = []

        self._phenomena: Dict[str, BoboPhenomenon] = {}

        for phenom in phenomena:
            if phenom.name not in self._phenomena:
                self._phenomena[phenom.name] = phenom
            else:
                raise BoboProducerError(
                    _EXC_PHENOM_NAME_DUP.format(phenom.name))

        self._gen_event_id: BoboGenEventID = gen_event_id
        self._gen_timestamp: BoboGenTimestamp = gen_timestamp
        self._max_size: int = max(0, max_size)
        self._queue: Queue[Tuple[BoboRunSerial, bool]] = Queue(self._max_size)

    def subscribe(self, subscriber: BoboProducerSubscriber):
        """
        :param subscriber: Subscriber to Producer data.
        """
        with self._lock:
            if subscriber not in self._subscribers:
                self._subscribers.append(subscriber)

    def update(self) -> bool:
        """
        :return: `True` if an internal update occurred; `False` otherwise.
        """
        with self._lock:
            if self._closed:
                return False

            if not self._queue.empty():
                event, local = self._queue.get_nowait()
                self._handle_completed_run(event, local)
                return True

            return False

    def close(self) -> None:
        """
        Closes the Producer.
        """
        with self._lock:
            self._closed = True

    def is_closed(self) -> bool:
        """
        :return: `True` if Producer is closed; `False` otherwise.
        """
        with self._lock:
            return self._closed

    def _handle_completed_run(
            self,
            runserial: BoboRunSerial,
            local: bool
    ) -> None:
        """
        :param runserial: The run to handle.
        :param local: `True` if the run was completed locally;
            `False` if the run was completed on a
            remote (distributed) instance.
        """
        if runserial.phenomenon_name not in self._phenomena:
            raise BoboProducerError(runserial.phenomenon_name)

        phenom: BoboPhenomenon = self._phenomena[runserial.phenomenon_name]

        event_complex = BoboEventComplex(
            event_id=self._gen_event_id.generate(),
            timestamp=self._gen_timestamp.generate(),
            data=phenom.datagen(phenom, runserial.history)
            if phenom.datagen is not None else None,
            phenomenon_name=runserial.phenomenon_name,
            pattern_name=runserial.pattern_name,
            history=runserial.history)

        for subscriber in self._subscribers:
            subscriber.on_producer_update(event=event_complex, local=local)

    def on_decider_update(
            self,
            completed: List[BoboRunSerial],
            halted: List[BoboRunSerial],
            updated: List[BoboRunSerial],
            local: bool
    ) -> None:
        """
        :param completed: Completed runs.
        :param halted: Halted runs.
        :param updated: Updated runs.
        :param local: `True` if the Decider update occurred locally;
            `False` if the update occurred on a remote (distributed) instance.
        """
        with self._lock:
            if self._closed:
                return

            for run in completed:
                if not self._queue.full():
                    self._queue.put((run, local))
                else:
                    raise BoboProducerError(
                        _EXC_QUEUE_FULL.format(self._max_size))

    def size(self) -> int:
        """
        :return: Queue size.
        """
        with self._lock:
            return self._queue.qsize()



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
A predicate tested against an event that must evaluate to `True` for the
event to be accepted as being part of the manifestation of a complex event.
"""

from abc import ABC, abstractmethod
from inspect import signature
from types import MethodType
from typing import Callable

from bobocep import BoboError
from bobocep.cep.event import BoboEvent, BoboHistory

EXC_INVALID_PARAM = "call must have {} parameters, found {}"
LEN_PARAM_CALL = 2


class BoboPredicateError(BoboError):
    """
    A predicate error.
    """


class BoboPredicate(ABC):
    """
    A predicate that evaluates to either True or False.
    """

    @abstractmethod
    def evaluate(self,
                 event: BoboEvent,
                 history: BoboHistory) -> bool:
        """
        Evaluates the predicate.

        :param event: An event.
        :type event: BoboEvent

        :param history: A history of events.
        :type history: BoboHistory

        :return: True if the predicate evaluates to True,
                 False otherwise.
        :rtype: bool
        """


class BoboPredicateCall(BoboPredicate):
    """
    A predicate that evaluates using a custom function or method
    (i.e. a 'callable').
    """

    def __init__(self, call: Callable):
        """
        :param call: The callable to use for evaluating the predicate.
        """
        super().__init__()

        len_param_call = len(signature(call).parameters)

        if len_param_call != LEN_PARAM_CALL:
            raise BoboPredicateError(
                EXC_INVALID_PARAM.format(LEN_PARAM_CALL, len_param_call))

        self._call = call

        # Prevent garbage collection of object if callable is a method.
        self._obj = call.__self__ if isinstance(call, MethodType) else None

    def evaluate(self, event: BoboEvent, history: BoboHistory) -> bool:
        """
        :param event: The event used for evaluation.
        :param history: The history of currently accepted events.
        :return: `True` if predicate is satisfied; `False` otherwise.
        """
        return self._call(event, history)


class BoboPredicateCallType(BoboPredicateCall):
    """
    A predicate that evaluates using a custom function or method after
    first checking whether the event data is an instance of a given type.
    If it is not, then a cast to the type can be attempted and a **copy of
    the event** is passed with its data cast to the type.

    **Note**: the copy is only used **within the predicate callable**.
    The original event remains in use elsewhere.
    """

    def __init__(self,
                 call: Callable,
                 dtype: type,
                 subtype: bool = True,
                 cast: bool = True):
        """
        :param call: The callable to use for evaluating the predicate.
        :param dtype: The data type to use for evaluation.
        :param subtype: If `True`, the event's data can be a subtype of
            the type specified in `dtype`.
            If `False`, it must be exactly the type in `dtype`.
        :param cast: If `True`, and if the event's data is not the expected
            type, then an attempt is made to cast it to `dtype`.
            If `False`, no attempt is made to cast the event's data.
        """
        super().__init__(call=call)

        self._dtype: type = dtype
        self._subtype: bool = subtype
        self._cast: bool = cast

    def evaluate(self, event: BoboEvent, history: BoboHistory) -> bool:
        """
        :param event: The event used for evaluation.
        :param history: The history of currently accepted events.
        :return: `True` if predicate is satisfied; `False` otherwise.
        """
        ok_type: bool = True

        if self._subtype:
            # Checking type and subtypes
            if not isinstance(event.data, self._dtype):
                ok_type = False
        else:
            # Checking exact type only
            if not type(event.data) == self._dtype:
                ok_type = False

        if not ok_type:
            if self._cast:
                # Type does not match, attempting cast
                try:
                    event = event.cast(self._dtype)

                except (TypeError, ValueError):
                    # Cast failed
                    return False
            else:
                # Not attempting cast
                return False

        # Type match or successful cast
        return self._call(event, history)



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Action definitions.
"""

from abc import ABC, abstractmethod
from typing import Tuple, Any

from bobocep import BoboError
from bobocep.cep.event import BoboEventComplex

_EXC_NAME_LEN = "name must have a length greater than 0"


class BoboActionError(BoboError):
    """
    An action error.
    """


class BoboAction(ABC):
    """
    An action.
    """

    def __init__(self, name: str, *args, **kwargs):
        """
        :param name: The action name.
        :param args: Action arguments.
        :param kwargs: Action keyword arguments.
        """
        super().__init__()

        if len(name) == 0:
            raise BoboActionError(_EXC_NAME_LEN)

        self._name = name
        self._args = args
        self._kwargs = kwargs

    @property
    def name(self) -> str:
        """
        :return: Action name.
        """
        return self._name

    @abstractmethod
    def execute(self, event: BoboEventComplex) -> Tuple[bool, Any]:
        """
        :param event: The complex event that triggered action.
        :return: Whether the action execution was successful, and
            any additional data.
        """



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Phenomenon imports.
"""

from bobocep.cep.phenom.pattern.builder import BoboPatternBuilder, \
    BoboPatternBuilderError
from bobocep.cep.phenom.pattern.pattern import BoboPattern, \
    BoboPatternBlock, BoboPatternError, BoboPatternBlockError
from bobocep.cep.phenom.pattern.predicate import BoboPredicate, \
    BoboPredicateError, BoboPredicateCall, BoboPredicateCallType
from bobocep.cep.phenom.phenom import BoboPhenomenon, \
    BoboPhenomenonError



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Abstract event.
"""

from abc import ABC, abstractmethod
from typing import Any

from bobocep import BoboError
from bobocep.bobocep import BoboJSONable

_EXC_ID_LEN = "event ID must have a length greater than 0"


class BoboEventError(BoboError):
    """
    An event error.
    """


class BoboEvent(BoboJSONable, ABC):
    """
    An abstract event.
    """

    EVENT_TYPE = "event_type"
    EVENT_ID = "event_id"
    TIMESTAMP = "timestamp"
    DATA = "data"

    def __init__(self,
                 event_id: str,
                 timestamp: int,
                 data: Any):
        """
        :param event_id: The event ID.
        :param timestamp: The event timestamp.
        :param data: The event data.

        :raises BoboEventError: If length of event ID is equal to 0.
        """
        super().__init__()

        if len(event_id) == 0:
            raise BoboEventError(_EXC_ID_LEN)

        self._event_id: str = event_id
        self._timestamp: int = timestamp
        self._data: Any = data

    @abstractmethod
    def cast(self, dtype: type) -> 'BoboEvent':
        """
        :param dtype: The type to which the event's data is cast.
        :return: A new BoboEvent instance with its data cast to `dtype`
            and all other properties identical to the original event.
        """

    @property
    def event_id(self) -> str:
        """
        Get event ID.
        """
        return self._event_id

    @property
    def timestamp(self) -> int:
        """
        Get event timestamp.
        """
        return self._timestamp

    @property
    def data(self) -> Any:
        """
        Get event data.
        """
        return self._data



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Generates BoboEvent objects.
"""

from abc import ABC, abstractmethod
from threading import RLock
from types import MethodType
from typing import Optional, Callable

from bobocep.cep.event import BoboEvent, BoboEventSimple
from bobocep.cep.gen.timestamp import BoboGenTimestamp, BoboGenTimestampEpoch


class BoboGenEvent(ABC):
    """
    An event generator.
    """

    @abstractmethod
    def maybe_generate(self, event_id: str) -> Optional[BoboEvent]:
        """
        :param event_id: The event ID to use in the generated event.
        :return: Either a generated BoboEvent instance or None.
        """


class BoboGenEventNone(BoboGenEvent):
    """
    An event generator that always returns None.
    """

    def maybe_generate(self, event_id: str) -> Optional[BoboEvent]:
        """
        :param event_id: An event ID (always ignored).
        :return: None.
        """
        return None


class BoboGenEventTime(BoboGenEvent):
    """
    An event generator that returns a BoboEventSimple time event if a given
    amount of time has elapsed. If the time has not elapsed, None is returned
    instead.
    """

    def __init__(self,
                 millis: int,
                 datagen: Optional[Callable] = None,
                 gen_timestamp: Optional[BoboGenTimestamp] = None,
                 from_now: bool = True,
                 tz=None):
        """
        :param millis: Milliseconds between event generations. Once the
            millisecond timer has elapsed, an event can be generated.
        :param datagen: Datagen to use in generated event.
        :param gen_timestamp: Custom timestamp generator.
            If None, a BoboGenTimestampEpoch instance is used to generate
            timestamps.
        :param from_now: If `True`, sets the millisecond timer to the current
            time; otherwise, the timer is set to 0.
        :param tz: Timezone data.
        """
        super().__init__()
        self._lock: RLock = RLock()

        self._millis: int = millis
        self._datagen: Optional[Callable] = datagen
        self._gen_ts_internal: BoboGenTimestampEpoch = BoboGenTimestampEpoch()
        self._tz = tz
        self._last: int = self._gen_ts_internal.generate() if from_now else 0
        self._gen_ts_event: BoboGenTimestamp = gen_timestamp \
            if gen_timestamp is not None else self._gen_ts_internal

        if self._datagen is not None:
            # Prevent garbage collection of object if callable is a method
            self._obj = datagen.__self__ \
                if isinstance(datagen, MethodType) else None

    def maybe_generate(self, event_id: str) -> Optional[BoboEvent]:
        """
        :param event_id: An event ID.
        :return: A BoboEventSimple instance if one is due to be generated,
            or `None` if not.
        """
        with self._lock:
            now = self._gen_ts_internal.generate()
            if (now - self._last) > self._millis:
                self._last = now
                return BoboEventSimple(
                    event_id=event_id,
                    timestamp=self._gen_ts_event.generate(),
                    data=self._datagen()
                    if self._datagen is not None else None)
            else:
                return None



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Generates BoboEvent IDs.
"""

from abc import ABC, abstractmethod
from threading import RLock
from time import time
from typing import Optional


class BoboGenEventID(ABC):
    """
    An event ID generator.
    """

    @abstractmethod
    def generate(self) -> str:
        """
        :return: Generated event ID.
        """


class BoboGenEventIDUnique(BoboGenEventID):
    """
    An event ID generator that always generates a unique, non-repeating ID.
    """

    def __init__(self, urn: Optional[str] = None):
        """
        :param urn: A URN to prefix before the generated event ID (optional).
        """
        super().__init__()
        self._lock: RLock = RLock()

        self._last: int = 0
        self._count: int = 0
        self._urn: Optional[str] = urn

    def generate(self) -> str:
        """
        :return: A generated event ID.
        """
        with self._lock:
            now: int = int(time())

            if now == self._last:
                self._count += 1
            else:
                self._count = 0
                self._last = now

            if self._urn is not None:
                return "{}_{}_{}".format(self._urn, now, self._count)
            else:
                return "{}_{}".format(now, self._count)



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Encryption for distributed `BoboCEP`.
"""

from abc import ABC, abstractmethod

from bobocep.dist.dist import BoboDistributedError


class BoboDistributedCryptoError(BoboDistributedError):
    """
    A distributed crypto error.
    """


class BoboDistributedCrypto(ABC):
    """
    Abstract class for distributed crypto.
    """

    @abstractmethod
    def encrypt(self, msg_str: str) -> bytes:
        """
        :param msg_str: Message to encrypt.
        :return: Encrypted message.
        """

    @abstractmethod
    def decrypt(self, msg_bytes: bytes) -> str:
        """
        :param msg_bytes: Message to decrypt.
        :return: Decrypted message.
        """

    @abstractmethod
    def end_bytes(self) -> bytes:
        """
        :return: Bytes used to signify the end of every encrypted payload.
        """

    @abstractmethod
    def min_length(self) -> int:
        """
        :return: The minimum possible length of an encrypted payload,
            in number of bytes.
        """



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
CEP engine.
"""

from threading import RLock

from bobocep import BoboError
from bobocep.cep.engine.decider.decider import BoboDecider
from bobocep.cep.engine.forwarder.forwarder import BoboForwarder
from bobocep.cep.engine.producer.producer import BoboProducer
from bobocep.cep.engine.receiver.receiver import BoboReceiver

_EXC_TIMES_REC = "receiver times must be greater than or equal to 0"
_EXC_TIMES_DEC = "decider times must be greater than or equal to 0"
_EXC_TIMES_PRO = "producer times must be greater than or equal to 0"
_EXC_TIMES_FOR = "forwarder times must be greater than or equal to 0"


class BoboEngineError(BoboError):
    """
    An engine error.
    """


class BoboEngine:
    """
    The engine for complex event processing.
    """

    def __init__(self,
                 receiver: BoboReceiver,
                 decider: BoboDecider,
                 producer: BoboProducer,
                 forwarder: BoboForwarder,
                 times_receiver: int = 0,
                 times_decider: int = 0,
                 times_producer: int = 0,
                 times_forwarder: int = 0,
                 early_stop: bool = True):
        """
        :param receiver: The receiver task.
        :param decider: The decider task.
        :param producer: The producer task.
        :param forwarder: The forwarder task.
        :param times_receiver: The number of times to run the receiver until
            moving to the decider task. A value of `0` runs the receiver
            indefinitely until it no longer performs an update of its
            internal state.
        :param times_decider: The number of times to run the decider until
            moving to the producer task. A value of `0` runs the decider
            indefinitely until it no longer performs an update of its
            internal state.
        :param times_producer: The number of times to run the producer until
            moving to the forwarder task. A value of `0` runs the producer
            indefinitely until it no longer performs an update of its
            internal state.
        :param times_forwarder: The number of times to run the forwarder until
            moving to the receiver task. A value of `0` runs the forwarder
            indefinitely until it no longer performs an update of its
            internal state.
        :param early_stop: If `times_*` is greater than `0`, it will always
            run the task for the set number of times, even if the task does
            not update. Setting `early_stop` to `True` stops early if no task
            update occurs.
        """
        super().__init__()
        self._lock: RLock = RLock()
        self._closed: bool = False

        if times_receiver < 0:
            raise BoboEngineError(_EXC_TIMES_REC)

        if times_decider < 0:
            raise BoboEngineError(_EXC_TIMES_DEC)

        if times_producer < 0:
            raise BoboEngineError(_EXC_TIMES_PRO)

        if times_forwarder < 0:
            raise BoboEngineError(_EXC_TIMES_FOR)

        self._times_receiver: int = times_receiver
        self._times_decider: int = times_decider
        self._times_producer: int = times_producer
        self._times_forwarder: int = times_forwarder
        self._early_stop: bool = early_stop

        self._receiver: BoboReceiver = receiver
        self._decider: BoboDecider = decider
        self._producer: BoboProducer = producer
        self._forwarder: BoboForwarder = forwarder

        self._receiver.subscribe(decider)
        self._decider.subscribe(producer)
        self._producer.subscribe(forwarder)
        self._producer.subscribe(receiver)
        self._forwarder.subscribe(receiver)

    @property
    def receiver(self) -> BoboReceiver:
        """
        Get receiver task.
        """
        return self._receiver

    @property
    def decider(self) -> BoboDecider:
        """
        Get decider task.
        """
        return self._decider

    @property
    def producer(self) -> BoboProducer:
        """
        Get producer task.
        """
        return self._producer

    @property
    def forwarder(self) -> BoboForwarder:
        """
        Get forwarder task.
        """
        return self._forwarder

    def close(self) -> None:
        """
        Closes the engine.
        """
        with self._lock:
            self._closed = True

            self._receiver.close()
            self._decider.close()
            self._producer.close()
            self._forwarder.close()

    def is_closed(self) -> bool:
        """
        :return: `True` if engine is set to close; `False` otherwise.
        """
        with self._lock:
            return self._closed

    def run(self) -> None:
        """
        Runs the engine. This is a blocking operation.
        """
        while True:
            with self._lock:
                if self._closed:
                    return

                self.update()

    def update(self) -> bool:
        """
        Performs a single update from the receiver, to the decider, to the
        producer, and finally to the forwarder.

        :return: `True` if engine is not set to close; `False` otherwise.
        """
        with self._lock:
            if self._closed:
                return False

            for task, times in [
                (self._receiver, self._times_receiver),
                (self._decider, self._times_decider),
                (self._producer, self._times_producer),
                (self._forwarder, self._times_forwarder)
            ]:
                if times == 0:
                    while task.update():
                        pass
                else:
                    for i in range(times):
                        if not task.update() and self._early_stop:
                            break

            return not self._closed



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Complex event.
"""

from json import dumps, loads
from typing import Any

from bobocep.cep.event.event import BoboEventError, BoboEvent
from bobocep.cep.event.history import BoboHistory

_EXC_PRO_LEN = "phenomenon name must have a length greater than 0"
_EXC_PAT_LEN = "pattern name must have a length greater than 0"


class BoboEventComplex(BoboEvent):
    """
    A complex event.
    """

    TYPE_COMPLEX = "type_complex"

    PHENOMENON_NAME = "phenomenon_name"
    PATTERN_NAME = "pattern_name"
    HISTORY = "history"

    def __init__(self,
                 event_id: str,
                 timestamp: int,
                 data: Any,
                 phenomenon_name: str,
                 pattern_name: str,
                 history: BoboHistory):
        """
        :param event_id: The event ID.
        :param timestamp: The event timestamp.
        :param data: The event data.
        :param phenomenon_name: The phenomenon name.
        :param pattern_name: The pattern name.
        :param history: The history of events.

        :raises BoboEventError: If length of phenomenon name is equal to 0.
        :raises BoboEventError: If length of pattern name is equal to 0.
        """
        super().__init__(
            event_id=event_id,
            timestamp=timestamp,
            data=data)

        if len(phenomenon_name) == 0:
            raise BoboEventError(_EXC_PRO_LEN)

        if len(pattern_name) == 0:
            raise BoboEventError(_EXC_PAT_LEN)

        self._phenomenon_name: str = phenomenon_name
        self._pattern_name: str = pattern_name
        self._history: BoboHistory = history

    def cast(self, dtype: type) -> 'BoboEventComplex':
        """
        :param dtype: The type to which the event's data is cast.
        :return: A new BoboEventComplex instance with its data cast to `dtype`
            and all other properties identical to the original event.
        """
        return BoboEventComplex(
            event_id=self._event_id,
            timestamp=self._timestamp,
            data=dtype(self._data),
            phenomenon_name=self._phenomenon_name,
            pattern_name=self._pattern_name,
            history=self._history
        )

    @property
    def phenomenon_name(self) -> str:
        """
        :return: Phenomenon name.
        """
        return self._phenomenon_name

    @property
    def pattern_name(self) -> str:
        """
        :return: Pattern name.
        """
        return self._pattern_name

    @property
    def history(self) -> BoboHistory:
        """
        :return: Event history.
        """
        return self._history

    def to_json_dict(self) -> dict:
        """
        :return: A JSON `dict` representation of the event.
        """
        return {
            self.EVENT_TYPE: self.TYPE_COMPLEX,
            self.EVENT_ID: self.event_id,
            self.TIMESTAMP: self.timestamp,
            self.DATA: self.data,
            self.PHENOMENON_NAME: self.phenomenon_name,
            self.PATTERN_NAME: self.pattern_name,
            self.HISTORY: self.history
        }

    def to_json_str(self) -> str:
        """
        :return: A JSON `str` representation of the event.
        """
        return dumps(self.to_json_dict(), default=lambda o: o.to_json_str())

    @staticmethod
    def from_json_str(j: str) -> 'BoboEventComplex':
        """
        :param j: A JSON `str` representation of the event.

        :return: A new instance of the event type.
        """
        return BoboEventComplex.from_json_dict(loads(j))

    @staticmethod
    def from_json_dict(d: dict) -> 'BoboEventComplex':
        """
        :param d: A JSON `dict` representation of the event.

        :return: A new instance of the event type.
        """
        return BoboEventComplex(
            event_id=d[BoboEventComplex.EVENT_ID],
            timestamp=d[BoboEventComplex.TIMESTAMP],
            data=d[BoboEventComplex.DATA],
            phenomenon_name=d[BoboEventComplex.PHENOMENON_NAME],
            pattern_name=d[BoboEventComplex.PATTERN_NAME],
            history=BoboHistory.from_json_str(d[BoboEventComplex.HISTORY])
        )

    def __str__(self) -> str:
        """
        :return: A JSON `str` representation of the event.
        """
        return self.to_json_str()



# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.
"""
Distributed publish-subscribe classes.
"""

from abc import ABC, abstractmethod
from typing import List

from bobocep.cep.engine.decider.runserial import BoboRunSerial


class BoboDistributedSubscriber(ABC):
    """A distributed subscriber interface."""

    @abstractmethod
    def on_distributed_update(
            self,
            completed: List[BoboRunSerial],
            halted: List[BoboRunSerial],
            updated: List[BoboRunSerial]) -> None:
        """
        :param completed: Completed runs.
        :param halted: Halted runs.
        :param updated: Updated runs.
        """


class BoboDistributedPublisher(ABC):
    """A distributed publisher interface."""

    @abstractmethod
    def subscribe(self, subscriber: BoboDistributedSubscriber) -> None:
        """
        :param subscriber: Subscriber to distributed.
        """
